/****************************************************************************
 *
 * KBruch2: Klasse KBruch korrekt abgeleitet von Klasse Bruch
 *
 *  - Beispielprogramm zum Buch "Objektorientiertes Programmieren in C++",
 *    von Nicolai Josuttis, erschienen bei Addison-Wesley, ISBN 3-89319-637-4
 *
 *  - Abschnitt 4.2.6: Vererbung richtig angewendet
 *  
 ****************************************************************************/

/****************************************************************************
 *
 * Copyright 1994 by Nicolai Josuttis and Addison-Wesley.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software for personal
 * and educational use is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the names of Addison-Wesley or the authors are not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Addison-Wesley and the authors make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * ADDISON-WESLEY AND THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL ADDISON-WESLEY OR THE AUTHORS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 ****************************************************************************/

// da die meisten Compiler "bool" (noch) nicht koennen (siehe Seite 43):
typedef int bool;
#define true 1
#define false 0



/****************************************************************************
 *
 *	Bruch.h
 *
 ****************************************************************************/

#ifndef _Bruch_h
#define _Bruch_h

/* allgemeine Header-Dateien
 */
#include <stdlib.h>
#include <iostream.h>

/* Klasse Bruch
 *  - zur Vererbung geeignet
 */
class Bruch {

  protected:
    int zaehler;
    int nenner;

  public:
    /* Default-Konstruktor, Konstruktor aus Zaehler und
     * Konstruktor aus Zaehler und Nenner
     */
    Bruch (int = 0, int = 1);

    // Multiplikation (globale Friend-Funktion)
    friend Bruch operator * (const Bruch&, const Bruch&);

    // multiplikative Zuweisung
    virtual const Bruch& operator *= (const Bruch&);
    // Vergleich (globale Friend-Funktion)
    friend bool operator < (const Bruch&, const Bruch&);

    // Ein- und Ausgabe mit Streams
    virtual void printOn (ostream& = cout) const;
    virtual void scanFrom (istream& = cin);

    // Konvertierung nach double
    virtual double asDouble () const;

    // virtueller Destruktor (ohne Anweisungen)
    virtual ~Bruch () {
    }
};

/* operator *
 *  - globale Friend-Funktion und inline deklariert
 */
inline Bruch operator * (const Bruch& a, const Bruch& b)
{
    /* Zaehler und Nenner einfach multiplizieren
     *  - kuerzen sparen wir uns
     */
    return Bruch (a.zaehler * b.zaehler, a.nenner * b.nenner);
}

/* Standard-Ausgabeoperator
 *  - global ueberladen und inline definiert
 */
inline ostream& operator << (ostream& strm, const Bruch& b)
{
    b.printOn (strm);      // Elementfunktion zur Ausgabe aufrufen
    return strm;           // Stream zur Verkettung zurueckliefern
}

/* Standard-Eingabeoperator
 *  - global ueberladen und inline definiert
 */
inline istream& operator >> (istream& strm, Bruch& b)
{
    b.scanFrom (strm);     // Elementfunktion zur Eingabe aufrufen
    return strm;           // Stream zur Verkettung zurueckliefern
}
/* asDouble
 *  - zur expliziten Typumwandlung von Bruch nach double
 */
inline double Bruch::asDouble () const
{
    // Quotient aus Zaehler und Nenner zurueckliefern
    return (double)zaehler / (double)nenner;
}

#endif  /* _Bruch_h */


/****************************************************************************
 *
 *	Bruch.cc / Bruch.cpp
 *
 ****************************************************************************/

// Header-Datei der Klasse einbinden
//#include "Bruch.h"

/* Default-Konstruktor, Konstruktor aus ganzer Zahl,
 * Konstruktor aus Zaehler und Nenner
 *  - Default fuer z: 0
 *  - Default fuer n: 1
 */
Bruch::Bruch (int z, int n)
{
    /* Zaehler und Nenner, wie uebergeben, initialisieren
     *  - 0 als Nenner ist allerdings nicht erlaubt
     *  - ein negatives Vorzeichen des Nenners kommt in den Zaehler
     */
    if (n == 0) {
        // NEU: 
        cerr << "Fehler: Nenner ist 0" << endl;
        exit (1);
    }
    if (n < 0) {
        zaehler = -z;
        nenner  = -n;
    }
    else {
        zaehler = z;
        nenner  = n;
    }
}

/* operator *=
 */
const Bruch& Bruch::operator *= (const Bruch& b)
{
    zaehler *= b.zaehler;
    nenner  *= b.nenner;

    // Objekt (erster Operand) wird zurueckgeliefert
    return *this;
}

/* operator <
 */
bool operator < (const Bruch& a, const Bruch& b)
{
    // Da die Nenner nicht negativ sein koennen reicht:
    return a.zaehler * b.nenner < b.zaehler * a.nenner;
}

/* NEU: printOn
 *  - Bruch auf Stream strm (default: cout) ausgeben
 */
void Bruch::printOn (ostream& strm) const
{
    strm << zaehler << '/' << nenner;
}

/* NEU: scanFrom
 *  - Bruch von Stream strm (default: cin) einlesen
 */
void Bruch::scanFrom (istream& strm)
{
    int z, n;

    // Zaehler einlesen
    strm >> z;

    // optionales Trennzeichen '/' einlesen
    if (strm.peek() == '/') {
        strm.get();
    }

    // Nenner einlesen
    strm >> n;

    // Lesefehler ?
    if (! strm) {
        return;
    }

    // Nenner == 0 ?
    if (n == 0) {
        // Fail-Bit setzen
        strm.clear (strm.rdstate() | ios::failbit);
        return;
    }

    /* OK, eingelesene Zahlen zuweisen
     *  - ein negatives Vorzeichen des Nenners kommt in den Zaehler
     */
    if (n < 0) {
        zaehler = -z;
        nenner  = -n;
    }
    else {
        zaehler = z;
        nenner  = n;
    }
}


/****************************************************************************
 *
 *	KBruch.h
 *
 ****************************************************************************/

#ifndef _KBruch_h
#define _KBruch_h

// Header-Datei der Basisklasse
//#include "Bruch.h"

/* Klasse KBruch 
 *  - von Bruch abgeleitet
 *  - der Zugriff auf geerbte Komponenten wird nicht
 *    eingeschraenkt ("public" bleibt "public")
 *  - zur Weiter-Vererbung geeignet
 */
class KBruch : public Bruch {
  protected:
    bool kuerzbar;        // true: Bruch ist kuerzbar

    // groesster gemeinsamer Teiler von zaehler und nenner
    unsigned ggt() const;

  public:
    /* Default-Konstruktor, Konstruktor aus Zaehler
     * und Konstruktor aus Zaehler und Nenner
     *  - Parameter werden an Bruch-Konstruktor durchgereicht
     */
    KBruch (int z = 0, int n = 1) : Bruch(z,n) {
        kuerzbar = (ggt() > 1);
    }

    // multiplikative Zuweisung (NEU IMPLEMENTIERT)
    virtual const Bruch& operator*= (const Bruch&);

    // Eingabe mit Streams (NEU IMPLEMENTIERT)
    virtual void scanFrom (istream& = cin);

    // Bruch kuerzen
    virtual void kuerzen();

    // liefert, ob kuerzbar
    virtual bool istKuerzbar() const {
        return kuerzbar;
    }

    // virtueller Destruktor (ohne Anweisungen)
    virtual ~KBruch () {
    }
};

#endif    /* _KBruch_h */


/****************************************************************************
 *
 *	KBruch.cc / KBruch.cpp
 *
 ****************************************************************************/

// Header-Datei der eigenen Klasse einbinden
//#include "KBruch.h"
/* min
 *  - statische Hilfsfunktion fuer das Minimum zweier int's
 */
inline static int min (int a, int b) {
    return a < b ? a : b;
}

/* ggt
 *  - groesster gemeinsamer Teiler von zaehler und nenner
 */
unsigned KBruch::ggt () const
{
    if (zaehler == 0) {
        return nenner;
    }

    /* Groesste Zahl ermitteln, die sowohl Zaehler als auch
     * Nenner ohne Rest teilt
     */
    unsigned teiler = min(abs(zaehler),nenner);
    while (zaehler % teiler != 0  ||  nenner % teiler != 0) {
        teiler--;
    }
    return teiler;
}

/* kuerzen()
 */
void KBruch::kuerzen ()
{
    /* falls kuerzbar,
     *  zaehler und nenner durch ggt teilen
     */
    if (kuerzbar) {
        int teiler = ggt();

        zaehler /= teiler;
        nenner  /= teiler;

        kuerzbar = false;       // damit nicht mehr kuerzbar
    }
}
/* Operator *=
 *  - zum Ueberschreiben mit Typen der Basisklasse neu implementiert
 */
const Bruch& KBruch::operator*= (const Bruch& b)
{
    /* Implementierung der Basisklasse aufrufen
     *  - auf nichtoeffentliche Komponenten von b besteht kein Zugriff
     */
    Bruch::operator*= (b);

    // weiterhin gekuerzt ?
    if (!kuerzbar) {
        kuerzbar = (ggt() > 1);
    }

    return  *this;
}

/* scanFrom
 */
void KBruch::scanFrom (istream& strm)
{
    // scanFrom() der Basisklasse aufrufen
    Bruch::scanFrom (strm);

    // Kuerzbarkeit testen
    kuerzbar = (ggt() > 1);
}


/****************************************************************************
 *
 *	kbtest2.cc / kbtest2.cpp
 *
 ****************************************************************************/

// Header-Dateien
#include <iostream.h>
//#include "KBruch.h"
//#include "Bruch.h"

main()
{
    // KBruch deklarieren
    KBruch x (7,3);

    // Bruch mit Kehrwert von x deklarieren
    Bruch b (3,7);

    // Zeiger auf Bruch zeigt darauf
    Bruch* xp = &x;

    *xp *= b;        // PROBLEM: ruft Bruch::operator*=() auf

    // x ausgeben
    cout << x;
    cout << (x.istKuerzbar() ? " (kuerzbar)" : " (unkuerzbar)") << endl;

    cout << "Bruch eingeben (zaehler nenner): " << flush;

    cin >> x;        // PROBLEM: ruft indirekt Bruch::scanFrom() auf

    // x ausgeben
    cout << x;
    cout << (x.istKuerzbar() ? " (kuerzbar)" : " (unkuerzbar)") << endl;
}

