/****************************************************************************
 *
 * Stack1: Template-Klasse Stack mit dynamischer Groesse
 *
 *  - Beispielprogramm zum Buch "Objektorientiertes Programmieren in C++",
 *    von Nicolai Josuttis, erschienen bei Addison-Wesley, ISBN 3-89319-637-4
 *
 *  - Abschnitt 3.14.2: Template-Klassen
 *  
 ****************************************************************************/

/****************************************************************************
 *
 * Copyright 1994 by Nicolai Josuttis and Addison-Wesley.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software for personal
 * and educational use is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the names of Addison-Wesley or the authors are not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Addison-Wesley and the authors make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * ADDISON-WESLEY AND THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL ADDISON-WESLEY OR THE AUTHORS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 ****************************************************************************/

// da die meisten Compiler "bool" (noch) nicht koennen (siehe Seite 43):
typedef int bool;
#define true 1
#define false 0



/****************************************************************************
 *
 *	Bruch.h
 *
 ****************************************************************************/

#ifndef _Bruch_h
#define _Bruch_h

// Standard-Header-Dateien einbinden
#include <stdlib.h>
#include <iostream.h>

class Bruch {

  private:
    int zaehler;
    int nenner;

  public:
    /* Default-Konstruktor, Konstruktor aus Zaehler und
     * Konstruktor aus Zaehler und Nenner
     */
    Bruch (int = 0, int = 1);
    // Multiplikation
    Bruch operator * (const Bruch&) const;
    
    // multiplikative Zuweisung
    const Bruch& operator *= (const Bruch&);
    
    // Vergleich
    bool operator < (const Bruch&) const;

    // NEU: Ausgabe mit Streams
    void printOn (ostream& = cout) const;

    // NEU: Eingabe mit Streams
    void scanFrom (istream& = cin);
};

/* operator *
 *  - inline definiert
 */
inline Bruch Bruch::operator * (const Bruch& b) const
{
    /* Zaehler und Nenner einfach multiplizieren
     *  - kuerzen sparen wir uns
     */
    return Bruch (zaehler * b.zaehler, nenner * b.nenner);
}

/* NEU: Standard-Ausgabeoperator
 *  - global ueberladen und inline definiert
 */
inline ostream& operator << (ostream& strm, const Bruch& b)
{
    b.printOn (strm);      // Elementfunktion zur Ausgabe aufrufen
    return strm;           // Stream zur Verkettung zurueckliefern
}

/* NEU: Standard-Eingabeoperator
 *  - global ueberladen und inline definiert
 */
inline istream& operator >> (istream& strm, Bruch& b)
{
    b.scanFrom (strm);     // Elementfunktion zur Eingabe aufrufen
    return strm;           // Stream zur Verkettung zurueckliefern
}

#endif  /* _Bruch_h */


/****************************************************************************
 *
 *	Bruch.cc / Bruch.cpp
 *
 ****************************************************************************/

// Header-Datei der Klasse einbinden
//#include "Bruch.h"

/* Default-Konstruktor, Konstruktor aus ganzer Zahl,
 * Konstruktor aus Zaehler und Nenner
 *  - Default fuer z: 0
 *  - Default fuer n: 1
 */
Bruch::Bruch (int z, int n)
{
    /* Zaehler und Nenner, wie uebergeben, initialisieren
     *  - 0 als Nenner ist allerdings nicht erlaubt
     *  - ein negatives Vorzeichen des Nenners kommt in den Zaehler
     */
    if (n == 0) {
        // NEU: 
        cerr << "Fehler: Nenner ist 0" << endl;
        exit (1);
    }
    if (n < 0) {
        zaehler = -z;
        nenner  = -n;
    }
    else {
        zaehler = z;
        nenner  = n;
    }
}

/* operator *=
 */
const Bruch& Bruch::operator *= (const Bruch& b)
{
    zaehler *= b.zaehler;
    nenner  *= b.nenner;

    // Objekt (erster Operand) wird zurueckgeliefert
    return *this;
}

/* operator <
 */
bool Bruch::operator < (const Bruch& b) const
{
    // Da die Nenner nicht negativ sein koennen reicht:
    return zaehler * b.nenner < b.zaehler * nenner;
}

/* NEU: printOn
 *  - Bruch auf Stream strm (default: cout) ausgeben
 */
void Bruch::printOn (ostream& strm) const
{
    strm << zaehler << '/' << nenner;
}

/* NEU: scanFrom
 *  - Bruch von Stream strm (default: cin) einlesen
 */
void Bruch::scanFrom (istream& strm)
{
    int z, n;

    // Zaehler einlesen
    strm >> z;

    // optionales Trennzeichen '/' einlesen
    if (strm.peek() == '/') {
        strm.get();
    }

    // Nenner einlesen
    strm >> n;

    // Lesefehler ?
    if (! strm) {
        return;
    }

    // Nenner == 0 ?
    if (n == 0) {
        // Fail-Bit setzen
        strm.clear (strm.rdstate() | ios::failbit);
        return;
    }

    /* OK, eingelesene Zahlen zuweisen
     *  - ein negatives Vorzeichen des Nenners kommt in den Zaehler
     */
    if (n < 0) {
        zaehler = -z;
        nenner  = -n;
    }
    else {
        zaehler = z;
        nenner  = n;
    }
}


/****************************************************************************
 *
 *	Stack.h
 *
 ****************************************************************************/

template <class T>
class Stack {
  private:
    T*  elems;            // dynamisches Array von Elementen vom Typ T
    int size;             // aktuelle Groesse
    int numElems;         // Anzahl eingetragener Elemente

  public:
    Stack ();             // Konstruktor

    /* aufgrund dynamischer Komponenten:
     */
    Stack (const Stack<T>&);                     // Copy-Konstruktor
    const Stack<T>& operator= (const Stack<T>&); // Zuweisung
    ~Stack ();                                   // Destruktor

    void push (T&);       // Element einkellern
    T pop ();             // Element auskellern
};


/****************************************************************************
 *
 *	Stack.cc / Stack.cpp
 *
 ****************************************************************************/

// Header-Dateien
#include <stdlib.h>
#include <iostream.h>
//#include "Stack.h"

/* Konstruktor
 */
template <class T>
Stack<T>::Stack ()
{
    numElems = 0;               // zunaechst keine Elemente
    size = 10;                  // Startgroesse: 10
    elems = new T [size];       // Speicherplatz anlegen
}

/* Copy-Konstruktor
 */
template <class T>
Stack<T>::Stack (const Stack<T>& s)
{
    numElems = s.numElems;      // Anzahl Elemente uebernehmen
    size = numElems;            // das ist auch die Startgroesse
    elems = new T [size];       // Speicherplatz anlegen

    // Elemente kopieren
    for (int i=0; i<numElems; i++) {
        elems[i] = s.elems[i];
    }
}

/* Destruktor
 */
template <class T>
Stack<T>::~Stack ()
{
    delete [] elems;            // Speicherplatz freigeben
}

/* Zuweisung
 */
template <class T>
const Stack<T>& Stack<T>::operator= (const Stack<T>& s)
{
    // Zuweisung an sich selbst macht nichts
    if (&s == this) {
        return *this;
    }

    numElems = s.numElems;      // Anzahl Elemente uebernehmen

    // falls nicht genug Speicherplatz, vergroessern
    if (size < numElems) {
        delete [] elems;            // alten Speicherplatz freigeben
        size = numElems;            // Anzahl Elemente ist neue Groesse
        elems = new T [size];       // neuen Speicherplatz anlegen
    }

    // Elemente kopieren
    for (int i=0; i<numElems; i++) {
        elems[i] = s.elems[i];
    }

    return *this;
}

/* push
 */
template <class T>
void Stack<T>::push (T& x)
{
    // falls nicht genug Speicherplatz, vergroessern
    if (size <= numElems) {
        T* tmp = elems;             // bisherige Elemente merken    
        size *= 2;                  // Groesse verdoppeln
        elems = new T [size];       // neuen Speicherplatz anlegen
        // bisherige Elemente in neuen Speicherplatz kopieren
        for (int i=0; i<numElems; i++) {
            elems[i] = tmp[i];
        }
        delete [] tmp;              // alten Speicherplatz freigeben
    }

    elems[numElems++] = x;      // neues Element einkellern
}

/* pop
 */
template<class T>
T Stack<T>::pop ()
{
    // kein Element mehr vorhanden ?
    if (numElems == 0) {
        cerr << "Stack ist leer" << endl;
        exit (1);
    }

    return elems[--numElems];   // Element auskellern
}



/****************************************************************************
 *
 *	stest1.cc / stest1.cpp
 *
 ****************************************************************************/

#include <iostream.h>
//#include "Bruch.h"
//#include "Stack.h"

typedef Stack<Bruch> BruchStack;

main()
{
    BruchStack bruchStack;     // Stack fuer Brueche

    static Bruch brueche [5] = {Bruch(7,3),Bruch(1,4),Bruch(42),Bruch(7,333)};
    int i;

    for (i=0; i<5; i++) {
        cout << "push: " << brueche[i] << endl;
	bruchStack.push(brueche[i]);
    }
    
    for (i=0; i<5; i++) {
        cout << "pop: " << bruchStack.pop() << endl;
    }
}

